/*
 * Copyright (c) 2010, Noel Lemouel
 * All rights reserved. See LICENCE file for details.
 */

#include "sysutil.h"

static char cpio_ascii_magic[] = "070707";

struct cpio_bin_header{
	unsigned short magic;
	unsigned short dev;
	unsigned short ino;
	unsigned short mode;
	unsigned short uid;
	unsigned short gid;
	unsigned short nlink;
	unsigned short rdev;
	unsigned long mtime;
	unsigned short namesize;
	unsigned long filesize;
};

struct cpio_ascii_header{
	unsigned char magic[6];		/* 070707 */
	unsigned char dev[6];
	unsigned char ino[6];
	unsigned char mode[6];
	unsigned char uid[6];
	unsigned char gid[6];
	unsigned char nlink[6];
	unsigned char rdev[6];
	unsigned char mtime[11];
	unsigned char namesize[6];
	unsigned char filesize[11];
};

/* ascii octal to integer */
static int n_aotoint(const char* s, unsigned long* i, int n)
{
	*i=0;
	while(*s && n--){
    	if ((*s >= 0x30) && (*s <= 0x39)){
    		*i = ((*i)*8) + (*s-0x30);
			s++;
    	}
	}
	return 0;
}

static int cpio_header_ascii_to_bin(struct cpio_ascii_header* ascii_hdr, struct cpio_bin_header* bin_hdr)
{
	unsigned long i;

	if(strncmp((char*)ascii_hdr->magic, cpio_ascii_magic, 6))
	{
		printf("invalid magic\n");
		return -1;
	}

	n_aotoint((char*)ascii_hdr->dev, &i, 6);
	bin_hdr->dev = (unsigned short)i;
	n_aotoint((char*)ascii_hdr->ino, &i, 6);
	bin_hdr->ino = (unsigned short)i;
	n_aotoint((char*)ascii_hdr->mode, &i, 6);
	bin_hdr->mode = (unsigned short)i;
	n_aotoint((char*)ascii_hdr->uid, &i, 6);
	bin_hdr->uid = (unsigned short)i;
	n_aotoint((char*)ascii_hdr->gid, &i, 6);
	bin_hdr->gid = (unsigned short)i;
	n_aotoint((char*)ascii_hdr->nlink, &i, 6);
	bin_hdr->nlink = (unsigned short)i;
	n_aotoint((char*)ascii_hdr->rdev, &i, 6);
	bin_hdr->rdev = (unsigned short)i;
	n_aotoint((char*)ascii_hdr->mtime, &i, 11);
	bin_hdr->mtime = i;
	n_aotoint((char*)ascii_hdr->namesize, &i, 6);
	bin_hdr->namesize = (unsigned short)i;
	n_aotoint((char*)ascii_hdr->filesize, &i, 11);
	bin_hdr->filesize = i;

	return 0;
}

static void print_field(unsigned char* s, int n)
{
   while(n--)
      printf("%c", *s++);

   printf("\n");
}

/* returns the offset to the file __header__ matching the name */
/* returns 0 if the file exists / -1 if it doesn't */
int cpio_find(unsigned char* base, const char* name, unsigned char** offset)
{
	struct cpio_ascii_header* ascii_hdr;
	struct cpio_bin_header bin_hdr;

	unsigned char* rec_off;
	unsigned char* pr;

	*offset = 0;
	rec_off = 0;

	while(1)
	{
		printf("record_offset: 0x%08x\n", rec_off);

		pr = (unsigned char*)((unsigned int)base+(unsigned int)rec_off);
		ascii_hdr = (struct cpio_ascii_header*)pr;

		/* extract header info. */

		if(cpio_header_ascii_to_bin(ascii_hdr, &bin_hdr)) return -1;

		/* get the record name */

		pr += sizeof(struct cpio_ascii_header);

		if(!strcmp((char*)pr, name)){
			printf("cpio-mm: '%s' found at offset 0x%08x\n", name, rec_off);
			*offset = rec_off;
			return 0;
		}
		/* no more files in the archive */
		else if(!strcmp((char*)pr, "TRAILER!!!")){
			printf("EOF at offset: 0x%08x\n", rec_off);
			return -1;
		}
		/* move to next record */

		/* filename is "null terminated" */
		/* namesize includes this extra null char */

		rec_off += (sizeof(struct cpio_ascii_header) + bin_hdr.namesize + bin_hdr.filesize);

	}

/*
   print_field(ascii_hdr->magic, 6);
   print_field(ascii_hdr->dev, 6);
   print_field(ascii_hdr->ino, 6);
   print_field(ascii_hdr->mode, 6);
   print_field(ascii_hdr->uid, 6);
   print_field(ascii_hdr->gid, 6);
   print_field(ascii_hdr->nlink, 6);
   print_field(ascii_hdr->rdev, 6);
   print_field(ascii_hdr->mtime, 11);
   print_field(ascii_hdr->namesize, 6);
   print_field(ascii_hdr->filesize, 11);

   printf("%d\n", bin_hdr.dev);
   printf("%d\n", bin_hdr.ino);
   printf("%d\n", bin_hdr.mode);
   printf("%d\n", bin_hdr.uid);
   printf("%d\n", bin_hdr.gid);
   printf("%d\n", bin_hdr.nlink);
   printf("%d\n", bin_hdr.rdev);
   printf("%ld\n", bin_hdr.mtime);
   printf("%d\n", bin_hdr.namesize);
   printf("%ld\n", bin_hdr.filesize);
*/
}

/* returns the base address (unsigned char** address) and size of the file __data__ matching "name" */
/* returns file_size if the file exists / -1 if it does */
unsigned int cpio_info(unsigned char* base, const char* name, unsigned char** address)
{
	struct cpio_ascii_header* ascii_hdr;
	struct cpio_bin_header bin_hdr;

	if(cpio_find(base, name, address))
	{
		printf("cpio-mm: '%s' not found\n", name);
		return 0;
	}

	*address += (unsigned int)base;

	ascii_hdr = (struct cpio_ascii_header*)*address;

	if(cpio_header_ascii_to_bin(ascii_hdr, &bin_hdr)) return -1;

	*address += (unsigned int)(bin_hdr.namesize + sizeof(struct cpio_ascii_header));

	printf("cpio-mm: '%s' data start at 0x%08x\n", name, *address);
	printf("cpio-mm: '%s' filesize: %d\n", name, bin_hdr.filesize);

	return bin_hdr.filesize;
}

/* load "size" bytes from memory region located at "address" to the "buffer" memory */
unsigned int cpio_load(unsigned char* buffer, unsigned char* address, unsigned int size)
{
	unsigned int i;

	for(i=0; i<size; i++)
	{
		buffer[i]=address[i];
	}

	return 0;
}

